// Velg hvilket modem som blir brukt
#define SIMPLE_NB_MODEM_SIM7080

// Enbale DBG debug output to Serial Monitor for debug prints, if needed
#define SIMPLE_NB_DEBUG Serial

// Velg Serial port for AT interface
#define SerialAT Serial2

// inkluder SimpleNB bibliotek
#include <SimpleNBClient.h>
#include <WiFi.h>

#define BAUD_RATE 115200  // Baud rate to be used for communicating with the modem

#define USE_SSL           // Kommenter bort denne linjen hvis det ikke blir brukt SSL
#ifdef USE_SSL
const int port = 443;
#else
const int port = 80;
#endif

// Server konfigurasjon
const char host[] = "snodybde-sensor.vercel.app";
const char resourceInit[] = "/api/import/sensor/BACHELOR23_SECRET_KEY";
const char resourceRef[] = "/api/import/referance/BACHELOR23_SECRET_KEY";
const char resourceData[] = "/api/import/sensor/data/BACHELOR23_SECRET_KEY";

int initialiseringen = 1;


// Pin for sensor 
#define SNO_PIN 5
#define VIND_PIN 25
#define TEMP_EXT_PIN 34
#define TEMP_INT_PIN 35

// Definer tidsintervall for avlesning i millisekunder
#define WRITE_INTERVAL 600000 // 10 minutter
#define READ_INTERVALSNO 30000 // 30 sekunder

// Variabler for snøsensoren
float pulsLengde = 0;
int filtrertVerdi = 0;
unsigned long forrigeAvlesningSno = 0;

// Definer filterkonstanter
const float alpha = 0.1;
const float beta = 1 - alpha;

// Definer variabler for rolling median filter
const int numMalinger = 20;     // Antall målinger som skal inkluderes i filteret
int malinger[numMalinger];     // Array for lagring av målingene
int sorted[numMalinger];       // Array for sorterte målinger
int malingerIndex = 0;         // Index for å legge til nye målinger i arrayet

// Definer konstant for maksimal avstand mellom avlesninger
const int maxDeviation = 600;

// Variabler for antall rotasjoner og tid siden siste avlesning
int rotasjoner = 0;
int hastighet = 0;
unsigned long lastReadTime = 0;
int rotasjonsTilstand = HIGH;
int forrigeRotasjonsTilstand = HIGH;

// Variabler for debounce-funksjonen
unsigned long debounceDelay = 50; // Debounce-tiden i millisekunder
unsigned long lastDebounceTime = 0;

// Variabel som holder styr på om vindsensoren er høy
bool roterer = false;

//variabler for temperatur sensorene 
float TempExt = 0;
float TempInt = 0;



SimpleNB modem(SerialAT);

void vindMaling(){
  // Les av knappens tilstand
    rotasjonsTilstand = digitalRead(VIND_PIN);

    // Sjekk om det har gått lang nok tid siden forrige debounce
    if ((millis() - lastDebounceTime) > debounceDelay) {
      // Sjekk om vindsensorens tilstand har endret seg fra forrige avlesning
      if (rotasjonsTilstand != forrigeRotasjonsTilstand) {
       // Oppdater siste debounce-tidspunkt
        lastDebounceTime = millis();
      } else {
        // Sjekk om knappen er trykket ned
        if (rotasjonsTilstand == HIGH && !roterer) {
          rotasjoner++;
          DBG("Knappetrykk registrert!");
          roterer = true;
        }

       // Sjekk om knappen er sluppet opp
       if (rotasjonsTilstand == LOW && roterer) {
         roterer = false;
        }
      }
    }
    forrigeRotasjonsTilstand = rotasjonsTilstand;
}

void snoMaling(){
  // Les lengden på PWM-signalet
    unsigned long pulsLengde = pulseIn(SNO_PIN, HIGH);

    // Legg til målingen i arrayet hvis den ligger innenfor en definert avstand fra gjennomsnittet
    float avg = filtrertVerdi;
    if (abs(pulsLengde - avg) <= maxDeviation) {
      malinger[malingerIndex] = pulsLengde;
      malingerIndex = (malingerIndex + 1) % numMalinger;

      // Beregn gjennomsnittet av avlesningene
      int sum = 0;
      int count = 0;
      for (int i = 0; i < numMalinger; i++) {
        if (malinger[i] != 0) {
          sum += malinger[i];
          count++;
        }
      }
      avg = (float)sum / (float)count;
      Serial.print("gjennomsnittet:");
      Serial.println(avg);

      // Oppdater filtrertVerdi
      filtrertVerdi = alpha * avg + beta * filtrertVerdi;
      Serial.print("filtrertVerdi: ");
      Serial.println(filtrertVerdi);
    } else {
      // Hvis avlesningen er utenfor avstanden, kast den og ta en ny avlesning
      Serial.println("Ugyldig avlesning: " + String(pulsLengde));
      pulsLengde = pulseIn(SNO_PIN, HIGH);
    }

    // Skriv ut avlesningene for debugging
    for (int i = 0; i < numMalinger; i++) {
      Serial.print("Reading ");
      Serial.print(i);
      Serial.print(": ");
      Serial.println(malinger[i]);
    }
  DBG(pulsLengde);
  forrigeAvlesningSno = millis();
}

void vind(){
  // Skriv ut antall rotasjoner siden forrige avlesning
      DBG("Antall rotasjoner i siste periode: ");
      DBG(rotasjoner);
      float rotasjon = (rotasjoner/2);
      DBG(rotasjon); 
      hastighet = ((rotasjon/(WRITE_INTERVAL))*1000)*40.84;
      DBG(hastighet);

      // Nullstill telleren
      rotasjoner = 0;  
}

void tempMaling(){
  TempExt = -((3600-2*(analogRead(TEMP_EXT_PIN)))/53);
  TempInt = -((3600-2*(analogRead(TEMP_INT_PIN)))/53);
}

void initialisering(){
  SimpleNBClientSecure client(modem, 0);
  DBG("Connecting to", host, "port", port);
  if (!client.connect(host, port)) {
    DBG("... failed");
  } else {
    // Make a HTTP POST request via TCP:
    client.println("POST " + String(resourceInit) + " HTTP/1.1");
    client.println("Host: " + String(host));
    client.println("Content-Type: application/json");
    client.println("Connection: close");
    client.print("Content-Length: ");
    String payload = "{\"id\":\"SENSOR_FERDIG\", \"name\": \"Olaf Kynsveens veg\"}";
    client.println(payload.length());
    client.println();
    client.println(payload);
      
    // Wait for response to arrive
    uint32_t start = millis();
    while (client.connected() && !client.available() && millis() - start < 30000L) {
      delay(100);
    };

    // Read response data, the response data consists of headers, payload acho backed so can be quite long
    start = millis();
    char received[680] = {'\0'};  // adjust the length accordingly as per your payload size
    int read_chars = 0;
    while (client.connected() && millis() - start < 10000L) {
      while (client.available()) {
        received[read_chars] = client.read();
        received[read_chars + 1] = '\0';
        read_chars++;
        start = millis();
      }
    }
    DBG(received);
    DBG("Data received:", strlen(received), "characters");
    client.stop();
    initialiseringen = 0; 
  }
}

void referanse(){
  SimpleNBClientSecure client(modem, 0);
  DBG("Connecting to", host, "port", port);
  if (!client.connect(host, port)) {
    DBG("... failed");
  } else {
    // Make a HTTP POST request via TCP:
    client.println("POST " + String(resourceRef) + " HTTP/1.1");
    client.println("Host: " + String(host));
    client.println("Content-Type: application/json");
    client.println("Connection: close");
    client.print("Content-Length: ");
    String payload = "{\"sensor_id\":\"SENSOR_FERDIG\", \"referance_point\": " + String(filtrertVerdi) + "}";
    client.println(payload.length());
    client.println();
    client.println(payload);
      
    // Wait for response to arrive
    uint32_t start = millis();
    while (client.connected() && !client.available() && millis() - start < 30000L) {
      delay(100);
    };

    // Read response data, the response data consists of headers, payload acho backed so can be quite long
    start = millis();
    char received[680] = {'\0'};  // adjust the length accordingly as per your payload size
    int read_chars = 0;
    while (client.connected() && millis() - start < 10000L) {
      while (client.available()) {
        received[read_chars] = client.read();
        received[read_chars + 1] = '\0';
        read_chars++;
        start = millis();
      }
    }
    DBG(received);
    DBG("Data received:", strlen(received), "characters");
    client.stop();
    initialiseringen = 0; 
  }
}


void sendData(){
  SimpleNBClientSecure client(modem, 0);
  DBG("Connecting to", host, "port", port);
    if (!client.connect(host, port)) {
      DBG("... failed");
    } else {
      // Make a HTTP POST request via TCP:
      client.println("POST " + String(resourceData) + " HTTP/1.1");
      client.println("Host: " + String(host));
      client.println("Content-Type: application/json");
      client.println("Connection: close");
      client.print("Content-Length: ");
      String payload = "{\"sensor_id\": \"SENSOR_FERDIG\","
                "\"current_depth\": " + String(filtrertVerdi) + ","
                "\"internal_temperature\": " + String(TempInt) + ","
                "\"external_temperature\": " + String(TempExt) + ","
                "\"wind_speed\": " + String(hastighet) +
                "}";
      client.println(payload.length());
      client.println();
      client.println(payload);

      // Wait for response to arrive
      uint32_t start = millis();
      while (client.connected() && !client.available() && millis() - start < 30000L) {
        delay(100);
      };

      // Read response data, the response data consists of headers, payload acho backed so can be quite long
      start = millis();
      char received[680] = {'\0'};  // adjust the length accordingly as per your payload size
      int read_chars = 0;
      while (client.connected() && millis() - start < 10000L) {
        while (client.available()) {
          received[read_chars] = client.read();
          received[read_chars + 1] = '\0';
          read_chars++;
          start = millis();
        }
      }
      DBG(received);
      DBG("Data received:", strlen(received), "characters");

      client.stop();
    }

    DBG("Deactivate Data Network... ");
    if (!modem.deactivateDataNetwork()) {
      delay(1000);
      return;
    }

    DBG("Powering Down...");
    //turn off module with AT command
    modem.powerOff();
    DBG("Done!\n\n");

    lastReadTime = millis();
}

void setup() {
    Serial.begin(115200);
    pinMode(VIND_PIN, INPUT_PULLUP);
    WiFi.mode(WIFI_OFF);

    // Fyll opp målinger-arrayet med de første 30 verdiene
    for (int i = 0; i < numMalinger; i++) {
      malinger[i] = pulseIn(SNO_PIN, HIGH);
      filtrertVerdi += malinger[i];
      delay(10);
    }
    filtrertVerdi /= numMalinger;
}

void loop() {

  vindMaling();
  
  // Sjekk om det har gått minst 1 minutter siden sist avlesning
  unsigned long currentTimesno = millis();
  if (currentTimesno - forrigeAvlesningSno >= READ_INTERVALSNO){
    snoMaling();
  }

  // Sjekk om det har gått minst 5 minutter siden sist avlesning
  unsigned long currentTime = millis();
  if (currentTime - lastReadTime >= WRITE_INTERVAL) {
    vind();
    tempMaling();
      
    // Set modem baud rate
    SimpleNBBegin(SerialAT, BAUD_RATE);

    DBG("Initializing modem...");
    modem.init();
      
    DBG("Waiting for network registration...");
    if (!modem.waitForRegistration(60000L, true)) {
      delay(1000);
    return;  // network timeout, can't register to the mobile network, check your sim or carrier setup
  }
    
    DBG("Activate Data Network...");
    if (!modem.activateDataNetwork()) {
      delay(1000);
    return;   // can't connect to the data network, check your APN and sim setup
  }

    if (initialiseringen == 1) {
      initialisering();
      referanse();
    }
    sendData();
  
  } 
}
